/*********
  Rui Santos
  Complete instructions at https://RandomNerdTutorials.com/esp8266-nodemcu-plot-readings-charts-multiple/

  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files.
  The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
*********/

#include <Arduino.h>
#include <ESP8266WiFi.h>
#include <ESPAsyncTCP.h>
#include <ESPAsyncWebServer.h>
#include "LittleFS.h"
#include <Arduino_JSON.h>
#include <OneWire.h>
#include <DallasTemperature.h>
#include <CircularBuffer.h>
#include <WiFiUdp.h>
#include <NTPClient.h>

WiFiUDP ntpUDP;
NTPClient timeClient(ntpUDP);

CircularBuffer<float, 25> buffer_temp;
CircularBuffer<unsigned long, 25> buffer_time;

//Haldenstr
// Replace with your network credentials

const char* ssid = "muesst ihr";
const char* password = "eintragen";

// Create AsyncWebServer object on port 80
AsyncWebServer server(80);

// Create an Event Source on /events
AsyncEventSource events("/events");

// Json Variable to hold sensor readings
JSONVar readings;
JSONVar readings_buffer;

// Timer variables
unsigned long lastTime = 0;
unsigned long lastTime2 = 0;
unsigned long timerDelayMeas = 500;
unsigned long timerDelaySend = 20000;

// GPIO where the DS18B20 sensors are connected to
const int oneWireBus = 4;

// Setup a oneWire instance to communicate with OneWire devices (DS18B20)
OneWire oneWire(oneWireBus);

// Pass our oneWire reference to Dallas Temperature sensor
DallasTemperature sensors(&oneWire);

// Address of each sensor
DeviceAddress sensor1 = { 0x28, 0xD3, 0xFD, 0x2E, 0xA, 0x0, 0x0, 0xC6 };

// Get Sensor Readings and return JSON object
void getSensorReadings(){
  // Uhrzeit holen
  timeClient.update();
  unsigned long ntp_time = timeClient.getEpochTime() + (unsigned long)(3600); // + 1h UTC
  buffer_time.push(ntp_time);
  Serial.println(ntp_time);
 
  // Temperatur lesen
  sensors.requestTemperatures();
  float Temp = sensors.getTempC(sensor1);
  buffer_temp.push(Temp);
  Serial.println(Temp);
}


// Get Sensor Readings and return JSON object
String getSensorReadingsBuffer(){

  // kopiere alle Daten vom CircularBuffer in JSON Buffer
  for (int i=0; i<25; i++){
    readings_buffer[i]["sensor1"] = String(buffer_temp[i]);
    readings_buffer[i]["time1"] = (unsigned long)buffer_time[i];
  }
  
  // alles in ein Objekt packen
//  String jsonBuffer = JSON.stringify(readings);
  String jsonBuffer = JSON.stringify(readings_buffer);
  return jsonBuffer;
}


// Initialize LittleFS
void initLittleFS() {
  if (!LittleFS.begin()) {
    Serial.println("An error has occurred while mounting LittleFS");
  }
  else{
    Serial.println("LittleFS mounted successfully");
  }
}

// Initialize WiFi
void initWiFi() {
  WiFi.mode(WIFI_STA);
  WiFi.begin(ssid, password);
  Serial.print("Connecting to WiFi ..");
  while (WiFi.status() != WL_CONNECTED) {
    Serial.print('.');
    delay(500);
  }
  Serial.println(WiFi.localIP());
}

void setup() {
  // Serial port for debugging purposes
  Serial.begin(115200);
  Serial.println("");
  initWiFi();
  initLittleFS();

  // Web Server Root URL
  server.on("/", HTTP_GET, [](AsyncWebServerRequest *request){
    request->send(LittleFS, "/index.html", "text/html");
  });

  server.serveStatic("/", LittleFS, "/");

  // Request for the latest sensor readings
  server.on("/readings", HTTP_GET, [](AsyncWebServerRequest *request){
    lastTime2 = lastTime2 + timerDelaySend;
    request->send(200, "text/plain", "OK!");
  });

  events.onConnect([](AsyncEventSourceClient *client){
    if(client->lastId()){
      Serial.printf("Client reconnected! Last message ID that it got is: %u\n", client->lastId());
    }
    // send event with message "hello!", id current millis
    // and set reconnect delay to 1 second
    client->send("hello!", NULL, millis(), 1000);
    // Sendet sofort alle Daten, die im Buffer sind, damit nicht gewartet werden muss
    lastTime2 = millis(); 
    events.send("ping",NULL,millis());
    events.send(getSensorReadingsBuffer().c_str(),"new_readings" ,millis());
    Serial.println("Send Buffer");
  });

  server.addHandler(&events);

  timeClient.begin();

  // Start server
  server.begin();
}


void loop() {
  // Messschleife
  if ((millis() - lastTime) > timerDelayMeas) {
    lastTime = millis(); 
    getSensorReadings();
  }

    // Sendschleife
  if ((millis() - lastTime2) > timerDelaySend) {
    // Send Events to the client with the Sensor Readings Every 10 seconds
    lastTime2 = millis(); 
    events.send("ping",NULL,millis());
    events.send(getSensorReadingsBuffer().c_str(),"new_readings" ,millis());
    Serial.println("Send Buffer");
  }
  
}
